@@ -107,3 +107,29 @@ serd_byte_source_close(SerdByteSource* const source)
   memset(source, '\0', sizeof(*source));
   return SERD_SUCCESS;
 }
+
+static SerdStatus
+peek_check(SerdByteSource* const source, const uint8_t byte)
+{
+  return serd_byte_source_peek(source) == byte ? SERD_SUCCESS
+                                               : SERD_ERR_BAD_SYNTAX;
+}
+
+SerdStatus
+serd_byte_source_skip_bom(SerdByteSource* const source)
+{
+  SerdStatus st = SERD_SUCCESS;
+
+  if (serd_byte_source_peek(source) == 0xEF) {
+    if (!(st = serd_byte_source_advance(source)) &&
+        !(st = peek_check(source, 0xBB)) &&
+        !(st = serd_byte_source_advance(source)) &&
+        !(st = peek_check(source, 0xBF))) {
+      st = serd_byte_source_advance(source);
+    } else {
+      st = st > SERD_FAILURE ? st : SERD_ERR_BAD_SYNTAX;
+    }
+  }
+
+  return st;
+}
