@@ -1,10 +1,13 @@
-// Copyright 2011-2025 David Robillard <d@drobilla.net>
+// Copyright 2011-2026 David Robillard <d@drobilla.net>
 // SPDX-License-Identifier: ISC
 
+#include "ntriples.h"
+#include "read_utf8.h"
 #include "reader.h"
 #include "serd_internal.h"
 #include "string_utils.h"
 #include "try.h"
+#include "turtle.h"
 
 #include <serd/serd.h>
 
@@ -76,16 +79,8 @@ read_UCHAR(SerdReader* const reader, const Ref dest, uint32_t* const char_code)
   }
 
   // Determine the encoded size from the code point
-  unsigned size = 0;
-  if (code < 0x00000080) {
-    size = 1;
-  } else if (code < 0x00000800) {
-    size = 2;
-  } else if (code < 0x00010000) {
-    size = 3;
-  } else if (code < 0x00110000) {
-    size = 4;
-  } else {
+  const unsigned size = utf8_num_bytes_for_codepoint(code);
+  if (!size) {
     r_err(reader,
           SERD_ERR_BAD_SYNTAX,
           "unicode character 0x%X out of range\n",
@@ -154,77 +149,6 @@ read_ECHAR(SerdReader* const reader, const Ref dest, SerdNodeFlags* const flags)
   }
 }
 
-static SerdStatus
-bad_char(SerdReader* const reader, const char* const fmt, const uint8_t c)
-{
-  // Skip bytes until the next start byte
-  for (int b = peek_byte(reader); b != EOF && ((uint8_t)b & 0x80);) {
-    skip_byte(reader, b);
-    b = peek_byte(reader);
-  }
-
-  r_err(reader, SERD_ERR_BAD_SYNTAX, fmt, c);
-  return reader->strict ? SERD_ERR_BAD_SYNTAX : SERD_FAILURE;
-}
-
-static SerdStatus
-read_utf8_bytes(SerdReader* const reader,
-                uint8_t           bytes[4],
-                uint8_t* const    size,
-                const uint8_t     c)
-{
-  *size = utf8_num_bytes(c);
-  if (*size <= 1) {
-    return bad_char(reader, "invalid UTF-8 start 0x%X\n", c);
-  }
-
-  bytes[0] = c;
-  for (uint8_t i = 1U; i < *size; ++i) {
-    const int b = peek_byte(reader);
-    if (b == EOF || ((uint8_t)b & 0x80U) == 0U) {
-      return bad_char(reader, "invalid UTF-8 continuation 0x%X\n", (uint8_t)b);
-    }
-
-    bytes[i] = (uint8_t)eat_byte_safe(reader, b);
-  }
-
-  return SERD_SUCCESS;
-}
-
-static SerdStatus
-read_utf8_character(SerdReader* const reader, const Ref dest, const uint8_t c)
-{
-  uint8_t    size     = 0U;
-  uint8_t    bytes[4] = {0, 0, 0, 0};
-  SerdStatus st       = read_utf8_bytes(reader, bytes, &size, c);
-  if (st) {
-    push_bytes(reader, dest, replacement_char, 3);
-  } else {
-    push_bytes(reader, dest, bytes, size);
-  }
-
-  return st;
-}
-
-static SerdStatus
-read_utf8_code(SerdReader* const reader,
-               const Ref         dest,
-               uint32_t* const   code,
-               const uint8_t     c)
-{
-  uint8_t    size     = 0U;
-  uint8_t    bytes[4] = {0, 0, 0, 0};
-  SerdStatus st       = read_utf8_bytes(reader, bytes, &size, c);
-  if (st) {
-    push_bytes(reader, dest, replacement_char, 3);
-    return st;
-  }
-
-  push_bytes(reader, dest, bytes, size);
-  *code = parse_counted_utf8_char(bytes, size);
-  return st;
-}
-
 // Read one character (possibly multi-byte)
 // The first byte, c, has already been eaten by caller
 static SerdStatus
@@ -243,7 +167,7 @@ read_character(SerdReader* const    reader,
     return push_byte(reader, dest, c);
   }
 
-  return read_utf8_character(reader, dest, c);
+  return read_utf8_continuation(reader, dest, c);
 }
 
 // [10] comment ::= '#' ( [^#xA #xD] )*
@@ -396,7 +320,7 @@ read_STRING_LITERAL(SerdReader* const    reader,
     }
   }
 
-  return eat_byte_check(reader, q) ? SERD_SUCCESS : SERD_ERR_BAD_SYNTAX;
+  return eat_byte_check(reader, q);
 }
 
 static SerdStatus
@@ -433,17 +357,6 @@ read_String(SerdReader* const    reader,
   return read_STRING_LITERAL_LONG(reader, node, flags, (uint8_t)q1);
 }
 
-static bool
-is_PN_CHARS_BASE(const uint32_t c)
-{
-  return ((c >= 0x00C0 && c <= 0x00D6) || (c >= 0x00D8 && c <= 0x00F6) ||
-          (c >= 0x00F8 && c <= 0x02FF) || (c >= 0x0370 && c <= 0x037D) ||
-          (c >= 0x037F && c <= 0x1FFF) || (c >= 0x200C && c <= 0x200D) ||
-          (c >= 0x2070 && c <= 0x218F) || (c >= 0x2C00 && c <= 0x2FEF) ||
-          (c >= 0x3001 && c <= 0xD7FF) || (c >= 0xF900 && c <= 0xFDCF) ||
-          (c >= 0xFDF0 && c <= 0xFFFD) || (c >= 0x10000 && c <= 0xEFFFF));
-}
-
 static SerdStatus
 read_PN_CHARS_BASE(SerdReader* const reader, const Ref dest)
 {
@@ -459,23 +372,17 @@ read_PN_CHARS_BASE(SerdReader* const reader, const Ref dest)
     return SERD_FAILURE;
   }
 
-  skip_byte(reader, c);
-  read_utf8_code(reader, dest, &code, (uint8_t)c);
+  if ((st = read_utf8_code_point(reader, dest, &code, (uint8_t)c))) {
+    return st;
+  }
 
-  if (!is_PN_CHARS_BASE(code)) {
+  if (!is_PN_CHARS_BASE((int)code)) {
     st = r_err_char(reader, "name", (int)code);
   }
 
   return st;
 }
 
-static bool
-is_PN_CHARS(const uint32_t c)
-{
-  return (is_PN_CHARS_BASE(c) || c == 0xB7 || (c >= 0x0300 && c <= 0x036F) ||
-          (c >= 0x203F && c <= 0x2040));
-}
-
 static SerdStatus
 read_PN_CHARS(SerdReader* const reader, const Ref dest)
 {
@@ -491,10 +398,11 @@ read_PN_CHARS(SerdReader* const reader, const Ref dest)
     return SERD_FAILURE;
   }
 
-  skip_byte(reader, c);
-  TRY(st, read_utf8_code(reader, dest, &code, (uint8_t)c));
+  if ((st = read_utf8_code_point(reader, dest, &code, (uint8_t)c))) {
+    return st;
+  }
 
-  if (!is_PN_CHARS(code)) {
+  if (!is_PN_CHARS((int)code)) {
     st = r_err_char(reader, "name", (int)code);
   }
 
@@ -521,10 +429,9 @@ read_PN_LOCAL_ESC(SerdReader* const reader, const Ref dest)
   skip_byte(reader, '\\');
 
   const int c = peek_byte(reader);
-  return ((c == '!') || in_range(c, '#', '/') || (c == ';') || (c == '=') ||
-          (c == '?') || (c == '@') || (c == '_') || (c == '~'))
-           ? push_byte(reader, dest, eat_byte_safe(reader, c))
-           : r_err(reader, SERD_ERR_BAD_SYNTAX, "bad escape\n");
+
+  return is_PN_LOCAL_ESC(c) ? push_byte(reader, dest, eat_byte_safe(reader, c))
+                            : r_err(reader, SERD_ERR_BAD_SYNTAX, "bad escape");
 }
 
 static SerdStatus
@@ -679,7 +586,7 @@ read_IRIREF_scheme(SerdReader* const reader, const Ref dest)
 static SerdStatus
 read_IRIREF(SerdReader* const reader, Ref* const dest)
 {
-  if (!eat_byte_check(reader, '<')) {
+  if (eat_byte_check(reader, '<')) {
     return SERD_ERR_BAD_SYNTAX;
   }
 
@@ -739,12 +646,9 @@ read_IRIREF(SerdReader* const reader, Ref* const dest)
           push_byte(reader, *dest, c);
         }
       } else if (!(c & 0x80)) {
-        push_byte(reader, *dest, c);
-      } else if (read_utf8_character(reader, *dest, (uint8_t)c)) {
-        if (reader->strict) {
-          *dest = pop_node(reader, *dest);
-          return SERD_ERR_BAD_SYNTAX;
-        }
+        st = push_byte(reader, *dest, c);
+      } else {
+        st = read_utf8_continuation(reader, *dest, (uint8_t)c);
       }
     }
   }
@@ -883,8 +787,8 @@ read_literal(SerdReader* const    reader,
     }
   } else if (next == '^') {
     skip_byte(reader, '^');
-    if (!eat_byte_check(reader, '^')) {
-      return r_err(reader, SERD_ERR_BAD_SYNTAX, "expected '^'\n");
+    if (eat_byte_check(reader, '^')) {
+      return SERD_ERR_BAD_SYNTAX;
     }
 
     if ((st = read_iri(reader, datatype, ate_dot))) {
@@ -939,7 +843,7 @@ read_BLANK_NODE_LABEL(SerdReader* const reader,
                       bool* const       ate_dot)
 {
   skip_byte(reader, '_');
-  if (!eat_byte_check(reader, ':')) {
+  if (eat_byte_check(reader, ':')) {
     return SERD_ERR_BAD_SYNTAX;
   }
 
@@ -991,8 +895,7 @@ read_BLANK_NODE_LABEL(SerdReader* const reader,
 static Ref
 read_blankName(SerdReader* const reader)
 {
-  if (skip_byte(reader, '=') || eat_byte_check(reader, '=') != '=') {
-    r_err(reader, SERD_ERR_BAD_SYNTAX, "expected '='\n");
+  if (skip_byte(reader, '=') || eat_byte_check(reader, '=')) {
     return 0;
   }
 
@@ -1058,9 +961,7 @@ read_anon(SerdReader* const reader,
     *ctx.flags = old_flags;
   }
 
-  return st > SERD_FAILURE                      ? st
-         : (eat_byte_check(reader, ']') == ']') ? SERD_SUCCESS
-                                                : SERD_ERR_BAD_SYNTAX;
+  return st > SERD_FAILURE ? st : eat_byte_check(reader, ']');
 }
 
 // Read a "named" object: a boolean literal or a prefixed name
@@ -1106,10 +1007,6 @@ read_object(SerdReader* const  reader,
             const bool         emit,
             bool* const        ate_dot)
 {
-#ifndef NDEBUG
-  const size_t orig_stack_size = reader->stack.size;
-#endif
-
   SerdStatus st = SERD_FAILURE;
 
   bool      simple   = (ctx->subject != 0);
@@ -1182,9 +1079,6 @@ read_object(SerdReader* const  reader,
   pop_node(reader, lang);
   pop_node(reader, datatype);
   pop_node(reader, o);
-#ifndef NDEBUG
-  assert(reader->stack.size == orig_stack_size);
-#endif
   return st;
 }
 
@@ -1215,7 +1109,7 @@ read_predicateObjectList(SerdReader* const reader,
          !(st = read_objectList(reader, ctx, ate_dot))) {
     ctx.predicate = pop_node(reader, ctx.predicate);
     if (*ate_dot) {
-      return SERD_SUCCESS;
+      break;
     }
 
     bool ate_semi = false;
@@ -1256,12 +1150,8 @@ end_collection(SerdReader* const reader,
   pop_node(reader, n2);
   pop_node(reader, n1);
   *ctx.flags &= ~(unsigned)SERD_LIST_CONT;
-  if (!st) {
-    return (eat_byte_check(reader, ')') == ')') ? SERD_SUCCESS
-                                                : SERD_ERR_BAD_SYNTAX;
-  }
 
-  return st;
+  return st ? st : eat_byte_check(reader, ')');
 }
 
 static SerdStatus
@@ -1351,7 +1241,6 @@ read_subject(SerdReader* const reader,
   }
 
   if (ate_dot) {
-    pop_node(reader, *dest);
     return r_err(reader, SERD_ERR_BAD_SYNTAX, "subject ends with '.'\n");
   }
 
@@ -1366,7 +1255,7 @@ read_labelOrSubject(SerdReader* const reader, Ref* const dest)
   case '[':
     skip_byte(reader, '[');
     read_ws_star(reader);
-    if (!eat_byte_check(reader, ']')) {
+    if (eat_byte_check(reader, ']')) {
       return SERD_ERR_BAD_SYNTAX;
     }
     *dest = blank_id(reader);
@@ -1418,20 +1307,20 @@ read_base(SerdReader* const reader, const bool sparql, const bool token)
   Ref uri = 0;
   TRY(st, read_IRIREF(reader, &uri));
   if (reader->base_sink) {
-    TRY(st, reader->base_sink(reader->handle, deref(reader, uri)));
+    st = reader->base_sink(reader->handle, deref(reader, uri));
   }
   pop_node(reader, uri);
 
   read_ws_star(reader);
   if (!sparql) {
-    return eat_byte_check(reader, '.') ? SERD_SUCCESS : SERD_ERR_BAD_SYNTAX;
+    return eat_byte_check(reader, '.');
   }
 
   if (peek_byte(reader) == '.') {
     return r_err(reader, SERD_ERR_BAD_SYNTAX, "full stop after SPARQL BASE\n");
   }
 
-  return SERD_SUCCESS;
+  return st;
 }
 
 static SerdStatus
@@ -1443,28 +1332,30 @@ read_prefixID(SerdReader* const reader, const bool sparql, const bool token)
   }
 
   read_ws_star(reader);
+
   Ref  name    = push_node(reader, SERD_LITERAL, "", 0);
   bool ate_dot = false;
-  TRY_FAILING(st, read_PN_PREFIX(reader, name, &ate_dot));
-  if (ate_dot || eat_byte_check(reader, ':') != ':') {
+  st           = read_PN_PREFIX(reader, name, &ate_dot);
+  if (st > SERD_FAILURE || ate_dot || eat_byte_check(reader, ':')) {
     pop_node(reader, name);
     return r_err(reader, SERD_ERR_BAD_SYNTAX, "expected a prefix name\n");
   }
 
   read_ws_star(reader);
-  Ref uri = 0;
-  TRY(st, read_IRIREF(reader, &uri));
 
-  if (reader->prefix_sink) {
+  Ref uri = 0;
+  st      = read_IRIREF(reader, &uri);
+
+  if (!st && reader->prefix_sink) {
     st = reader->prefix_sink(
       reader->handle, deref(reader, name), deref(reader, uri));
   }
 
   pop_node(reader, uri);
   pop_node(reader, name);
-  if (!sparql) {
+  if (!st && !sparql) {
     read_ws_star(reader);
-    st = eat_byte_check(reader, '.') ? SERD_SUCCESS : SERD_ERR_BAD_SYNTAX;
+    st = eat_byte_check(reader, '.');
   }
 
   return st;
@@ -1498,39 +1389,45 @@ read_directive(SerdReader* const reader)
 static SerdStatus
 read_wrappedGraph(SerdReader* const reader, ReadContext* const ctx)
 {
-  if (!eat_byte_check(reader, '{')) {
+  SerdStatus st = SERD_SUCCESS;
+
+  if (eat_byte_check(reader, '{')) {
     return SERD_ERR_BAD_SYNTAX;
   }
 
   read_ws_star(reader);
-  while (peek_byte(reader) != '}') {
-    bool ate_dot  = false;
-    int  s_type   = 0;
-    ctx->subject  = 0;
-    SerdStatus st = read_subject(reader, *ctx, &ctx->subject, &s_type);
+  while ((st <= SERD_FAILURE) && peek_byte(reader) != '}') {
+    int s_type   = 0;
+    ctx->subject = 0;
+    st           = read_subject(reader, *ctx, &ctx->subject, &s_type);
     if (st) {
-      return r_err(reader, SERD_ERR_BAD_SYNTAX, "bad subject\n");
+      st = r_err(reader, SERD_ERR_BAD_SYNTAX, "bad subject\n");
     }
 
-    if ((st = read_triples(reader, *ctx, &ate_dot)) && s_type != '[') {
-      return r_err(reader, st, "bad predicate object list\n");
+    bool ate_dot = false;
+    if (!st && (st = read_triples(reader, *ctx, &ate_dot)) && s_type != '[') {
+      st = r_err(reader,
+                 st > SERD_FAILURE ? st : SERD_ERR_BAD_SYNTAX,
+                 "bad predicate object list\n");
     }
 
     ctx->subject = pop_node(reader, ctx->subject);
+    if (st <= SERD_FAILURE) {
+      read_ws_star(reader);
+      st = (peek_byte(reader) == '.') ? skip_byte(reader, '.') : SERD_SUCCESS;
+      read_ws_star(reader);
+    }
+  }
+
+  if (!st) {
+    skip_byte(reader, '}');
     read_ws_star(reader);
     if (peek_byte(reader) == '.') {
-      skip_byte(reader, '.');
+      return r_err(reader, SERD_ERR_BAD_SYNTAX, "graph followed by '.'\n");
     }
-    read_ws_star(reader);
   }
 
-  skip_byte(reader, '}');
-  read_ws_star(reader);
-  if (peek_byte(reader) == '.') {
-    return r_err(reader, SERD_ERR_BAD_SYNTAX, "graph followed by '.'\n");
-  }
-
-  return SERD_SUCCESS;
+  return st;
 }
 
 static bool
@@ -1557,10 +1454,6 @@ token_equals(SerdReader* const reader,
 SerdStatus
 read_n3_statement(SerdReader* const reader)
 {
-#ifndef NDEBUG
-  const size_t orig_stack_size = reader->stack.size;
-#endif
-
   SerdStatementFlags flags   = 0;
   ReadContext        ctx     = {0, 0, 0, 0, 0, 0, &flags};
   bool               ate_dot = false;
@@ -1578,20 +1471,26 @@ read_n3_statement(SerdReader* const reader)
       return r_err(
         reader, SERD_ERR_BAD_SYNTAX, "syntax does not support directives\n");
     }
-    TRY(st, read_directive(reader));
-    read_ws_star(reader);
+    if (!(st = read_directive(reader))) {
+      read_ws_star(reader);
+    }
     break;
   case '{':
     if (reader->syntax == SERD_TRIG) {
-      TRY(st, read_wrappedGraph(reader, &ctx));
-      read_ws_star(reader);
+      if (!(st = read_wrappedGraph(reader, &ctx))) {
+        read_ws_star(reader);
+      }
     } else {
       return r_err(
         reader, SERD_ERR_BAD_SYNTAX, "syntax does not support graphs\n");
     }
     break;
   default:
-    TRY_FAILING(st, read_subject(reader, ctx, &ctx.subject, &s_type));
+    st = read_subject(reader, ctx, &ctx.subject, &s_type);
+    if (st > SERD_FAILURE) {
+      ctx.subject = pop_node(reader, ctx.subject);
+      return st;
+    }
 
     if (token_equals(reader, ctx.subject, "base", 4)) {
       st = read_base(reader, true, false);
@@ -1600,49 +1499,47 @@ read_n3_statement(SerdReader* const reader)
     } else if (token_equals(reader, ctx.subject, "graph", 5)) {
       ctx.subject = pop_node(reader, ctx.subject);
       read_ws_star(reader);
-      TRY(st, read_labelOrSubject(reader, &ctx.graph));
-      read_ws_star(reader);
-      TRY(st, read_wrappedGraph(reader, &ctx));
-      pop_node(reader, ctx.graph);
-      ctx.graph = 0;
+      if (!(st = read_labelOrSubject(reader, &ctx.graph))) {
+        read_ws_star(reader);
+        st = read_wrappedGraph(reader, &ctx);
+      }
+      ctx.graph = pop_node(reader, ctx.graph);
       read_ws_star(reader);
     } else if (token_equals(reader, ctx.subject, "true", 4) ||
                token_equals(reader, ctx.subject, "false", 5)) {
-      return r_err(reader, SERD_ERR_BAD_SYNTAX, "expected subject\n");
+      st = r_err(reader, SERD_ERR_BAD_SYNTAX, "expected subject\n");
     } else if (read_ws_star(reader) && peek_byte(reader) == '{') {
       if (s_type == '(' || (s_type == '[' && !*ctx.flags)) {
+        ctx.subject = pop_node(reader, ctx.subject);
         return r_err(reader, SERD_ERR_BAD_SYNTAX, "invalid graph name\n");
       }
       ctx.graph   = ctx.subject;
-      ctx.subject = 0;
-      TRY(st, read_wrappedGraph(reader, &ctx));
-      pop_node(reader, ctx.graph);
+      ctx.subject = 0; // FIXME: remove?
+      st          = read_wrappedGraph(reader, &ctx);
+      ctx.graph   = pop_node(reader, ctx.graph);
       read_ws_star(reader);
     } else if ((st = read_triples(reader, ctx, &ate_dot))) {
       if (st == SERD_FAILURE && s_type == '[') {
+        ctx.subject = pop_node(reader, ctx.subject);
         return SERD_SUCCESS;
       }
 
       if (ate_dot && (reader->strict || (s_type != '('))) {
+        ctx.subject = pop_node(reader, ctx.subject);
         return r_err(
           reader, SERD_ERR_BAD_SYNTAX, "unexpected end of statement\n");
       }
 
-      return st > SERD_FAILURE ? st : SERD_ERR_BAD_SYNTAX;
+      st = st > SERD_FAILURE ? st : SERD_ERR_BAD_SYNTAX;
     } else if (!ate_dot) {
       read_ws_star(reader);
-      st = (eat_byte_check(reader, '.') == '.') ? SERD_SUCCESS
-                                                : SERD_ERR_BAD_SYNTAX;
+      st = eat_byte_check(reader, '.');
     }
 
     ctx.subject = pop_node(reader, ctx.subject);
     break;
   }
 
-#ifndef NDEBUG
-  assert(reader->stack.size == orig_stack_size);
-#endif
-
   return st;
 }
 
@@ -1697,40 +1594,45 @@ read_nquads_statement(SerdReader* const reader)
   }
 
   // subject predicate object
-  if ((st = read_subject(reader, ctx, &ctx.subject, &s_type)) ||
-      !read_ws_star(reader) || (st = read_IRIREF(reader, &ctx.predicate)) ||
-      !read_ws_star(reader) ||
-      (st = read_object(reader, &ctx, false, &ate_dot))) {
-    return st;
+  if (!(st = read_subject(reader, ctx, &ctx.subject, &s_type))) {
+    read_ws_star(reader);
+    if (!(st = read_IRIREF(reader, &ctx.predicate))) {
+      read_ws_star(reader);
+      st = read_object(reader, &ctx, false, &ate_dot);
+    }
   }
 
-  if (!ate_dot) { // graphLabel?
+  if (!st && !ate_dot) { // graphLabel?
     read_ws_star(reader);
     switch (peek_byte(reader)) {
     case '.':
       break;
     case '_':
-      TRY(st, read_BLANK_NODE_LABEL(reader, &ctx.graph, &ate_dot));
+      st = read_BLANK_NODE_LABEL(reader, &ctx.graph, &ate_dot);
       break;
     default:
-      TRY(st, read_IRIREF(reader, &ctx.graph));
+      st = read_IRIREF(reader, &ctx.graph);
     }
 
     // Terminating '.'
     read_ws_star(reader);
-    if (!ate_dot && !eat_byte_check(reader, '.')) {
-      return SERD_ERR_BAD_SYNTAX;
+    if (!ate_dot) {
+      st = eat_byte_check(reader, '.');
     }
   }
 
-  TRY(st, emit_statement(reader, ctx, ctx.object, ctx.datatype, ctx.lang));
+  if (!st) {
+    st = emit_statement(reader, ctx, ctx.object, ctx.datatype, ctx.lang);
+  }
 
   pop_node(reader, ctx.graph);
   pop_node(reader, ctx.lang);
   pop_node(reader, ctx.datatype);
   pop_node(reader, ctx.object);
+  pop_node(reader, ctx.predicate);
+  pop_node(reader, ctx.subject);
 
-  return SERD_SUCCESS;
+  return st;
 }
 
 SerdStatus
