@@ -36,9 +36,9 @@
 #define SERD_SRC_SERD_CONFIG_H
 
 // Define version unconditionally so a warning will catch a mismatch
-#define SERD_VERSION "0.32.6"
+#define SERD_VERSION "0.32.8"
 
-#if !defined(SERD_NO_DEFAULT_CONFIG)
+#ifndef SERD_NO_DEFAULT_CONFIG
 
 // We need unistd.h to check _POSIX_VERSION
 #  ifdef __has_include
@@ -50,7 +50,7 @@
 #  endif
 
 // Define SERD_POSIX_VERSION unconditionally for convenience
-#  if defined(_POSIX_VERSION)
+#  ifdef _POSIX_VERSION
 #    define SERD_POSIX_VERSION _POSIX_VERSION
 #  else
 #    define SERD_POSIX_VERSION 0
