@@ -10,6 +10,7 @@
 #include <assert.h>
 #include <stdint.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
 #define NS_EG "http://example.org/"
@@ -214,25 +215,37 @@ test_write_bad_anon_stack(void)
 static void
 test_strict_write(void)
 {
-  SerdEnv* const    env    = serd_env_new(NULL);
-  SerdWriter* const writer = serd_writer_new(
-    SERD_TURTLE, (SerdStyle)SERD_STYLE_STRICT, env, NULL, null_sink, NULL);
+  SerdEnv* const    env = serd_env_new(NULL);
+  SerdWriter* const writer =
+    serd_writer_new(SERD_TURTLE,
+                    (SerdStyle)(SERD_STYLE_STRICT | SERD_STYLE_ASCII),
+                    env,
+                    NULL,
+                    null_sink,
+                    NULL);
   assert(writer);
 
-  const uint8_t bad_str[] = {0xFF, 0x90, 'h', 'i', 0};
+  static const uint8_t bad_uri_buf[]   = {'f', 't', 'p', ':', 0xFF, 0x90, 0};
+  static const uint8_t bad_short_buf[] = {0xFF, 0x90, 'h', 'i', 0};
+  static const uint8_t bad_long_buf[]  = {'e', '\n', 'r', 0xFF, 0x90, 0};
 
-  SerdNode s = serd_node_from_string(SERD_URI, USTR(NS_EG "s"));
-  SerdNode p = serd_node_from_string(SERD_URI, USTR(NS_EG "p"));
-
-  SerdNode bad_lit = serd_node_from_string(SERD_LITERAL, bad_str);
-  SerdNode bad_uri = serd_node_from_string(SERD_URI, bad_str);
-
-  assert(serd_writer_write_statement(
-           writer, 0, NULL, &s, &p, &bad_lit, NULL, NULL) == SERD_ERR_BAD_TEXT);
+  SerdNode s             = serd_node_from_string(SERD_URI, USTR(NS_EG "s"));
+  SerdNode p             = serd_node_from_string(SERD_URI, USTR(NS_EG "p"));
+  SerdNode bad_uri       = serd_node_from_string(SERD_URI, bad_uri_buf);
+  SerdNode bad_short_lit = serd_node_from_string(SERD_LITERAL, bad_short_buf);
+  SerdNode bad_long_lit  = serd_node_from_string(SERD_LITERAL, bad_long_buf);
 
   assert(serd_writer_write_statement(
            writer, 0, NULL, &s, &p, &bad_uri, NULL, NULL) == SERD_ERR_BAD_TEXT);
 
+  assert(serd_writer_write_statement(
+           writer, 0, NULL, &s, &p, &bad_short_lit, NULL, NULL) ==
+         SERD_ERR_BAD_TEXT);
+
+  assert(serd_writer_write_statement(
+           writer, 0, NULL, &s, &p, &bad_long_lit, NULL, NULL) ==
+         SERD_ERR_BAD_TEXT);
+
   serd_writer_free(writer);
   serd_env_free(env);
 }
@@ -376,6 +389,79 @@ test_write_bad_statement(void)
   serd_env_free(env);
 }
 
+static void
+check_pname_escape(const char* const lname, const char* const expected)
+{
+  SerdEnv* const env   = serd_env_new(NULL);
+  SerdChunk      chunk = {NULL, 0};
+
+  SerdWriter* const writer = serd_writer_new(SERD_TURTLE,
+                                             (SerdStyle)SERD_STYLE_CURIED,
+                                             env,
+                                             NULL,
+                                             serd_chunk_sink,
+                                             &chunk);
+  assert(writer);
+
+  static const char* const prefix     = NS_EG;
+  const size_t             prefix_len = strlen(prefix);
+
+  serd_env_set_prefix_from_strings(env, USTR("eg"), USTR(NS_EG));
+
+  SerdNode s = serd_node_from_string(SERD_URI, USTR(NS_EG "s"));
+  SerdNode p = serd_node_from_string(SERD_URI, USTR(NS_EG "p"));
+
+  char* const uri = (char*)calloc(1, prefix_len + strlen(lname) + 1);
+  assert(uri);
+  memcpy(uri, prefix, prefix_len + 1);
+  memcpy(uri + prefix_len, lname, strlen(lname) + 1);
+
+  SerdNode o = serd_node_from_string(SERD_URI, USTR(uri));
+
+  assert(
+    !serd_writer_write_statement(writer, 0U, NULL, &s, &p, &o, NULL, NULL));
+  assert(!serd_writer_finish(writer));
+  free(uri);
+
+  uint8_t* const out = serd_chunk_sink_finish(&chunk);
+  assert(out);
+  assert(!strcmp((const char*)out, expected));
+  serd_free(out);
+
+  serd_writer_free(writer);
+  serd_env_free(env);
+}
+
+static void
+test_write_pname_escapes(void)
+{
+  // Check that '.' is escaped only at the start and end
+  check_pname_escape(".xyz", "eg:s\n\teg:p eg:\\.xyz .\n");
+  check_pname_escape("w.yz", "eg:s\n\teg:p eg:w.yz .\n");
+  check_pname_escape("wx.z", "eg:s\n\teg:p eg:wx.z .\n");
+  check_pname_escape("wxy.", "eg:s\n\teg:p eg:wxy\\. .\n");
+
+  // Check that ':' is not escaped anywhere
+  check_pname_escape(":xyz", "eg:s\n\teg:p eg::xyz .\n");
+  check_pname_escape("w:yz", "eg:s\n\teg:p eg:w:yz .\n");
+  check_pname_escape("wx:z", "eg:s\n\teg:p eg:wx:z .\n");
+  check_pname_escape("wxy:", "eg:s\n\teg:p eg:wxy: .\n");
+
+  // Check that special characters like '~' are escaped everywhere
+  check_pname_escape("~xyz", "eg:s\n\teg:p eg:\\~xyz .\n");
+  check_pname_escape("w~yz", "eg:s\n\teg:p eg:w\\~yz .\n");
+  check_pname_escape("wx~z", "eg:s\n\teg:p eg:wx\\~z .\n");
+  check_pname_escape("wxy~", "eg:s\n\teg:p eg:wxy\\~ .\n");
+
+  // Check that out of range multi-byte characters are escaped everywhere
+  static const char first_escape[] = {(char)0xC3U, (char)0xB7U, 'y', 'z', 0};
+  static const char mid_escape[]   = {'w', (char)0xC3U, (char)0xB7U, 'z', 0};
+  static const char last_escape[]  = {'w', 'x', (char)0xC3U, (char)0xB7U, 0};
+  check_pname_escape((const char*)first_escape, "eg:s\n\teg:p eg:%C3%B7yz .\n");
+  check_pname_escape((const char*)mid_escape, "eg:s\n\teg:p eg:w%C3%B7z .\n");
+  check_pname_escape((const char*)last_escape, "eg:s\n\teg:p eg:wx%C3%B7 .\n");
+}
+
 int
 main(void)
 {
@@ -388,6 +474,7 @@ main(void)
   test_chunk_sink();
   test_write_nothing_node();
   test_write_bad_statement();
+  test_write_pname_escapes();
 
   return 0;
 }
